#include "nav.h"
//#include <stdio.h>
#include <alloc.h>
#include <math.h>

// **************************************************************************
//				Die Navigationsklasse startet
// **************************************************************************

Navigation::Navigation()
{
char	far	*pointer;

	m_intWegGefunden = 0;            // Kein Weg gefunden setzen
	m_KartenZeiger= (char*) farmalloc( KARTENGROESSE * (KARTENGROESSE / 8) ); // Bitfeld

	if ( m_KartenZeiger != NULL)								// Konnte die Karte angelegt werden
		{
		pointer = m_KartenZeiger;
		while (pointer <= (m_KartenZeiger + (KARTENGROESSE * (KARTENGROESSE / 8) )))
			{
			*pointer=0;
			pointer++;
			};
		PathSucher.SetRows( KARTENGROESSE);
		PathSucher.SetData(m_KartenZeiger);
		}

	PathSucher.SetValid(Navigation::FeldFrei);
	PathSucher.SetCost(Navigation::Kosten);
};

// **************************************************************************
//				Die Navigationsklasse wird beendet
// **************************************************************************

Navigation::~Navigation()
{
	if(m_KartenZeiger != NULL )						// Besteht noch eine Karte im Speicher
		{
		farfree( m_KartenZeiger );						// Letzten Speicher wieder freigeben
		}
};

// **************************************************************************
//					Gib die Maske zurueck, welche das Zielfeld adressiert
// **************************************************************************

char	Navigation::BitMaske(int x)
{
char	byte;

	byte = 1 << (x % 8 );				// Schiebe Bit zur Zielposition
	return byte;
};

// **************************************************************************
//				Gib einen Zeiger auf das Byte im Bitfeld
// Die Karte besteht aus einem Bitfeld, ein Byte beinhaltet somit
// 8 Felder der Karte um Speicher zu sparen
// **************************************************************************

char far	*Navigation::BitPointer(int x, int y)
{
unsigned	long	lngOffset;
char	far	*pointer;

	pointer = m_KartenZeiger;					// Pointer von Basisadresse legen	
	lngOffset= y * (KARTENGROESSE / 8);		// Adresse des Zeilenbeginn
	lngOffset += x / 8;							// Spaltenoffset hinzufuegen
	pointer += lngOffset;                  // Zieladresse berechnen
	return (pointer);								// Rueckgabe des Zeiger auf das Zielbyte
};

// **************************************************************************
//                Fuege Hinderniss in die Karte ein
// **************************************************************************

void	Navigation::SetzeHindernis(int x, int y)
{
char far *pointer;
char	byte;

	pointer = BitPointer( x, y );             // Ermittle Speicherbyte
	byte    = BitMaske( x );      				// Ermittle Speicherbit
	*pointer |= byte;								  	// Hindernis setzen	
};

// **************************************************************************
//                Loesche ein Hinderniss in die Karte ein
// **************************************************************************

void	Navigation::LoescheHindernis(int x, int y)
{
char far *pointer;
char	byte;

	pointer = BitPointer( x, y);              // Ermittle Speicherbyte
	byte    = BitMaske( x );      				// Ermittle Speicherbit
	*pointer &= ~byte;								// Hindernis loeschen
};

// **************************************************************************
//       Ist an dieser Stelle ein Hindernis in der Karte
// **************************************************************************

int	Navigation::IstHindernis(int x, int y)
{
char far *pointer;
char	byte;
char	ergebnis;

	if( x>KARTENGROESSE ) return 1;				// Ausserhalb der Karte, ja
	if( y>KARTENGROESSE ) return 1;				// Ausserhalb der Karte, ja

	pointer = BitPointer( x, y);              // Ermittle Speicherbyte
	byte    = BitMaske( x );      				// Ermittle Speicherbit
	ergebnis = (char) *pointer & byte;					// Maske Speicherbit aus
	return ((int) ergebnis);
};

// **************************************************************************
//        Gehe zum Startpunkt
// **************************************************************************

_asNode *Navigation::WegBeginn(void)
{
  if( m_intWegGefunden )						// Wurde ein Weg gefunden
	 {                                           
	 return (PathSucher.GetFirstNode());
	 }
  else
	 return ( NULL );	            	  		// Keinen Weg gefunden
};

// **************************************************************************
//        Gib den naechste Koordinate heraus, zum Ziel
// **************************************************************************

_asNode *Navigation::WegWeiter(void)
{
  if( m_intWegGefunden )						// Wurde ein Weg gefunden
	 {                                           
	 return PathSucher.GetNextNode();
	 }
  else
	 return ( NULL );	            	  		// Keinen Weg gefunden
};

// **************************************************************************
//        Pruefe nach ob das Feld (die Zelle) betreten werden darf
// **************************************************************************

int Navigation::FeldFrei(int x, int y, void far *p) 
{
char far *pointer;
unsigned long	lngOffset;
char	byte;
char  buffer;

	if( x>KARTENGROESSE ) return 0;			// Ausserhalb der Karte, nein
	if( y>KARTENGROESSE ) return 0;			// Ausserhalb der Karte, nein

	pointer = (char *) p;						// Pointer von Basisadresse legen	
	lngOffset= y * (KARTENGROESSE / 8);  	// Adresse des Zeilenbeginn
	lngOffset += (x / 8);						// X Offset ermitteln
	pointer += lngOffset;                  // Zieladresse berechnen
	buffer = (char) *pointer;					// Wert laden								
	byte = 1 << (x % 8 );						// Schiebe Bit zur Zielposition

	if ( buffer & byte )							// Ist das Hindernisbit gesetzt
		return 0;                           // Feld nicht betreten
	else
		return 1;									// Feld betreten
}

// **************************************************************************
//                    Kosten fuer das betreten des Feldes
// **************************************************************************

int Navigation::Kosten(int x, int y, void *p)
{
	return ( 1 );
};

// **************************************************************************
//               Setze den Startpunkt des zu suchenden Weges
// **************************************************************************

void Navigation::StartPunkt(int x, int y)
{
	if (x > KARTENGROESSE) return;			// Bereichskontrolle x
	if (y > KARTENGROESSE)  return;			// Bereichskontrolle y
	
	m_StartPunkt.x=x;
	m_StartPunkt.y=y;
};

// **************************************************************************
//               Setze den Endpunkt des zu suchenden Weges
// **************************************************************************

void Navigation::EndPunkt(int x, int y)
{
	if (x > KARTENGROESSE) return;			// Bereichskontrolle x
	if (y > KARTENGROESSE)  return;			// Bereichskontrolle y

	m_EndPunkt.x=x;
	m_EndPunkt.y=y;
};

// **************************************************************************
// 	         			Suche jetzt einen Weg
// **************************************************************************

void Navigation::Find() 
{
	if( m_KartenZeiger != NULL )
	  {	
	  m_intWegGefunden = PathSucher.GeneratePath(m_StartPunkt.x,
																m_StartPunkt.y,
																m_EndPunkt.x,
																m_EndPunkt.y);
	  }
  else
	  m_intWegGefunden = 0;							 // Keine Karte, kein Weg
};

// **************************************************************************
//    Gib den internen Status zurueck, ob ein Weg gefunden wurde
// **************************************************************************

int  Navigation::WegGefunden( void )
{
	return m_intWegGefunden;
}
