//
//  Datei Path.cpp
//  Author  :	 R.Wenzel (Ursprung von Generation5.org)
// Bezzeichnung: A* Algorithmus
//

#include "Path.h"

// *************************************************************************
//       Die Klasse wird neu angelegt (init routine)
// *************************************************************************

CPathFinder::CPathFinder() 
{
	m_pOpen = m_pClosed = NULL;
	m_pStack = NULL;
}

// *************************************************************************
//         Die Klasse wird beendet
// *************************************************************************

CPathFinder::~CPathFinder() 
{
	ClearNodes();
}

// *************************************************************************
//          Loesche alle Knoten
// *************************************************************************

void CPathFinder::ClearNodes()
{
	_asNode *temp = NULL;

	if (m_pOpen)
		{
		while (m_pOpen)
			{
			temp = m_pOpen->next;
			delete m_pOpen;
			m_pOpen = temp;
			}
		}

	if (m_pClosed)
		{
		while (m_pClosed) 
			{
			temp = m_pClosed->next;
			delete m_pClosed;
			m_pClosed = temp;
			}
		}
}

// *************************************************************************
//                Suche den Path auf der Karte
// *************************************************************************

int CPathFinder::GeneratePath(int sx, int sy, int dx, int dy)
{
	ClearNodes();
	
	m_iSX = sx;								// Setze Start Koordinaten
	m_iSY = sy;
	m_iDX = dx;                      // Setze End Koordinaten
	m_iDY = dy;
	m_iDNum = Coord2Num(dx,dy);

	_asNode *temp = new _asNode(sx, sy);
	_asNode *best;

	temp->g = 0;
	temp->h = (dx-sx)*(dx-sx) + (dy-sy)*(dy-sy);
	temp->f = temp->g + temp->h;
	temp->number = Coord2Num(sx, sy);

	m_pOpen = temp;
	while (1)
		{
		if (!(best = GetBest())) return 0;

		if (best->number == m_iDNum) break;

		CreateChildren(best);
		};
	
	m_pBest = best;

	return 1;
};

// *************************************************************************
//           Gib den ersten Eintrag des Weges zurueck
// *************************************************************************

_asNode *CPathFinder::GetFirstNode()
{
_asNode	*pointer;
_asNode	*temp;

	temp = GetBestNode();					// Besten Knoten bestimmen
	temp->next = NULL;						// Ende der Liste
	while ( temp->parent ) 				  	// So lange es Eintraege gibt
		  {
		  pointer = temp;						// Merke mir das "child"			  
		  temp = temp->parent;				// Ermittl den vorherigen Eintrag
		  temp->next = pointer;				// Verlinke  
		  }
	m_WalkList = temp;						// Merke mir den ersten Punkt
	return m_WalkList;
}

// *************************************************************************
//					Gib den naechsten Link zurueck
// *************************************************************************

_asNode *CPathFinder::GetNextNode()
{
	if(m_WalkList->next)								// Gib des noch Folgelinks
		{
		m_WalkList = m_WalkList->next;			// naechsten Eintrag nehmen
		return m_WalkList;								// gib den aktuellen Link
		}
	else
		return NULL;								  // Ende der Liste	
}

// *************************************************************************
//                 Gib den End Path des Weges zurueck
// *************************************************************************

_asNode *CPathFinder::GetBest()
{
	if (!m_pOpen) return NULL;				// wurde berhaupt ein Weg gefunden

	_asNode *temp = m_pOpen;				// Ermittle den ersten Knoten der open Liste
	_asNode *temp2 = m_pClosed;			// ebenso der close Liste

	m_pOpen = temp->next;					

	if (temp2)
		{
		m_pClosed = temp;
		m_pClosed->next = temp2;
		}
	else 
		{
		m_pClosed = temp;
		m_pClosed->next = NULL;
		}

	return temp;                        // Gefundenen Link zurueckgeben
}

// *************************************************************************
//      Suche einen weiteren Knoten ab und lege einen link an
// *************************************************************************

void CPathFinder::CreateChildren(_asNode *node)
{
	int x = node->x;
	int y = node->y;

	for (int i=-1;i<2;i++) 
		{
		for (int j=-1;j<2;j++) 
			{
			if (i == 0 && j == 0 || !udValid(x+i, y+j, m_pCBData)) 
				continue;

			LinkChild(node, x+i, y+j);
			}
		}
}

// *************************************************************************
//
// *************************************************************************

void CPathFinder::LinkChild(_asNode *node, int x, int y)
{
	int g = node->g + udCost(x,y,m_pCBData);
	int num = Coord2Num(x,y);

	_asNode *check = NULL;

	if ((check = CheckList(m_pOpen, num)) != 0) 
		{
		node->children[node->numchildren++] = check;
		
		// A better route found, so update
		// the node and variables accordingly.

		if (g < check->g) 
			{
			check->parent = node;
			check->g = g;
			check->f = g + check->h;
			}
		}
	else
		if ((check = CheckList(m_pClosed, num))  != 0 ) 
			{
			node->children[node->numchildren++] = check;

			if (g < check->g) 
				{
				check->parent = node;
				check->g = g;
				check->f = g + check->h;

				// The fun part...
				UpdateParents(check);
				}
			}
		else
			{
			_asNode *newnode = new _asNode(x,y);
			newnode->parent = node;
			newnode->g = g;
			newnode->h = (x-m_iDX)*(x-m_iDX) + (y-m_iDY)*(y-m_iDY);
			newnode->f = newnode->g + newnode->h;
			newnode->number = Coord2Num(x,y);

			AddToOpen(newnode);

			node->children[node->numchildren++] = newnode;
			}
}

// *************************************************************************
//
// *************************************************************************

_asNode *CPathFinder::CheckList(_asNode *node, int num) 
{
	while (node)
		{
		if (node->number == num) 
			return node;
		node = node->next;
		}

	return NULL;
}

// *************************************************************************
//                  sort with respect to f
// *************************************************************************

void CPathFinder::AddToOpen(_asNode *addnode) 
{
	_asNode *node = m_pOpen;
	_asNode *prev = NULL;

	if (!m_pOpen) 
		{
		m_pOpen = addnode;
		return;
		}

	while(node) 
		{
		if (addnode->f > node->f) 
			{
			prev = node;
			node = node->next;
			}
		else
			{
			if (prev)
				{
				prev->next = addnode;
				addnode->next = node;
				}
			else
				{
				_asNode *temp = m_pOpen;

				m_pOpen = addnode;
				m_pOpen->next = temp;
				}

			return;
			}
		 }

	prev->next = addnode;
}

// *************************************************************************
// 						Have to propagate the changes down.
// *************************************************************************

void CPathFinder::UpdateParents(_asNode *node)
{
	int g = node->g, c = node->numchildren;

	_asNode *kid = NULL;
	for (int i=0;i<c;i++)
		{
		kid = node->children[i];
		if (g+1 < kid->g)
			{
			kid->g = g+1;
			kid->f = kid->g + kid->h;
			kid->parent = node;
			
			Push(kid);
			}
		}

	_asNode *parent;
	while (m_pStack) 
		{
		parent = Pop();
		c = parent->numchildren;
		for (int i=0;i<c;i++)
			{
			kid = parent->children[i];
			
			if (parent->g+1 < kid->g) 
				{
				kid->g = parent->g + udCost(kid->x, kid->y,m_pCBData);
				kid->f = kid->g + kid->h;
				kid->parent = parent;

				Push(kid);
				}
			}
		}
}

// *************************************************************************
//             Knoten auf den Stack legen
// *************************************************************************

void CPathFinder::Push(_asNode *node) 
{

	if (!m_pStack)                           	// Existiert ein Stack ?
		{                                    	// Nein, dann ersten anlegen
		m_pStack = new _asStack;             	// Neuen Knoten anlegen
		m_pStack->data = node;                 // Aus den Stack legen
		m_pStack->next = NULL;                 // Kein Link zum naechsten 
		}
	else
		{
		_asStack *temp = new _asStack;         // Neues Element anlegen

		temp->data = node;                     // Daten ablegen
		temp->next = m_pStack;                 // im Stack einhaengen
		m_pStack = temp;
		}
}

// *************************************************************************
//             Knoten vom Listen Stack holen
// *************************************************************************

_asNode *CPathFinder::Pop() 
{
	_asNode *data = m_pStack->data;
	_asStack *temp = m_pStack;
	m_pStack = temp->next;
	delete temp;
	return data;
}
